/* util.c   by Michael Thorpe   2010-06-11 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

static obj *freelist=0;
#ifdef DEBUG
unsigned long total_objects=0;
static obj *firstobj=0;
#endif

static objtype *base_objtype=0;

obj *new_object(objtype *type) {
	obj *o;

	if(freelist) {
		o=freelist;
		freelist=o->rcnf.nextfree;
		if(o->type->dropobj)
			(o->type->dropobj)(o);
	} else {
		o=(obj *)calloc(1,sizeof(obj));
		if(!o)
			return(throwoom());
	}
	o->type=type;
	o->rcnf.refcount=1;
#ifdef DEBUG
#if DEBUG > 1
	fprintf(stderr,"new_object(%s)=%p\n",type->name,o);
#endif
	total_objects++;
	o->prev=0;
	o->next=firstobj;
	if(firstobj)
		firstobj->prev=o;
	firstobj=o;
#endif
	return(o);
}

void drop_object(obj *o) {
#if DEBUG > 1
	fprintf(stderr,"drop_object(%p)\n",o);
#endif
	if(o->type->predropobj)
		(o->type->predropobj)(o);
#ifdef DEBUG
	if(o->type->dropobj)
		(o->type->dropobj)(o);
	total_objects--;
	if(o->prev)
		o->prev->next=o->next;
	else
		firstobj=o->next;
	if(o->next)
		o->next->prev=o->prev;
	memset(o,'o',sizeof(obj));
#else
	o->rcnf.nextfree=freelist;
	freelist=o;
#endif
}

#ifdef DEBUG
obj *real_incref(obj *o,const char *file,int line) {
#if DEBUG > 2
	fprintf(stderr,"%s:%d: INCREF(%p)\n",file,line,o);
#endif
	if(o->rcnf.refcount>10000000)
		*(char *)0=0;
	o->rcnf.refcount++;
	return(o);
}

void real_decref(obj *o,const char *file,int line) {
#if DEBUG > 2
	fprintf(stderr,"%s:%d: DECREF(%p)\n",file,line,o);
#endif
	if(!--o->rcnf.refcount)
		drop_object(o);
}
#endif

int register_objtype(objtype *type) {
	if(type->nameobj)
		return(1); /* Already initialized? */
	type->nameobj=new_word(type->name,strlen(type->name));
	if(!type->nameobj)
		return(1);
	if(!type->tostring)
		return(1);
	type->next=base_objtype;
	base_objtype=type;
	if(type->init) {
		if((*type->init)()) {
			fprintf(stderr,"registering objtype %s failed\n",type->name);
			return(1);
		}
	}
	return(0);
}

#ifdef DEBUG
void show_remaining_objects() {
	obj *o;

	for(o=firstobj;o;o=o->next) {
		if((obj *)0x6f6f6f6f==o->next) {
			fprintf(stderr,"%p: dropped object found!\n",o);
			break;
		}
		fprintf(stderr,"%p: <%lu:%s>\n",o,o->rcnf.refcount,o->type->name);
if(&word_objtype==o->type)fprintf(stderr,"\t\"%.*s\"\n",(int)o->value.word.extra->len,o->value.word.extra->data);
	}
}
#endif

int clean_up() {
	obj *o;

	while(base_objtype) {
		if(base_objtype->exit)
			(*base_objtype->exit)();
		DECREF(base_objtype->nameobj);
		base_objtype=base_objtype->next;
	}
	while(freelist) {
		o=freelist;
		freelist=o->rcnf.nextfree;
		if(o->type->dropobj)
			(o->type->dropobj)(o);
		free(o);
	}
#ifdef DEBUG
	fflush(stdout);
	if(total_objects) {
		fprintf(stderr,"Total objects left over: %lu\n",total_objects);
		show_remaining_objects();
	}
	return(!!total_objects);
#else
	return(0);
#endif
}
